﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;
using HeuristicLab.Core;

namespace GradientBoostedTrees {
  public class RelativeErrorLoss : ILossFunction {
    public double GetLoss(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetEnum = target.GetEnumerator();
      var predEnum = pred.GetEnumerator();
      var weightEnum = weight.GetEnumerator();

      double s = 0;
      while (targetEnum.MoveNext() & predEnum.MoveNext() & weightEnum.MoveNext()) {
        double res = targetEnum.Current - predEnum.Current;
        s += weightEnum.Current * Math.Abs(res) * Math.Abs(1.0 / targetEnum.Current);
      }
      if (targetEnum.MoveNext() | predEnum.MoveNext() | weightEnum.MoveNext())
        throw new ArgumentException("target, pred and weight have differing lengths");

      return s;
    }

    public IEnumerable<double> GetLossGradient(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetEnum = target.GetEnumerator();
      var predEnum = pred.GetEnumerator();
      var weightEnum = weight.GetEnumerator();

      while (targetEnum.MoveNext() & predEnum.MoveNext() & weightEnum.MoveNext()) {
        var res = targetEnum.Current - predEnum.Current;
        if (res > 0) yield return weightEnum.Current * 1.0 / Math.Abs(targetEnum.Current);
        else if (res < 0) yield return -weightEnum.Current * 1.0 / Math.Abs(targetEnum.Current);
        else yield return 0.0;
      }
      if (targetEnum.MoveNext() | predEnum.MoveNext() | weightEnum.MoveNext())
        throw new ArgumentException("target, pred and weight have differing lengths");
    }

    public LineSearchFunc GetLineSearchFunc(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetArr = target.ToArray();
      var predArr = pred.ToArray();
      var weightArr = weight.ToArray();
      Debug.Assert(weightArr.All(w => w.IsAlmost(1.0)));

      if (targetArr.Length != predArr.Length || predArr.Length != weightArr.Length)
        throw new ArgumentException("target, pred and weight have differing lengths");

      // line search for abs error
      LineSearchFunc lineSearch = (idx, startIdx, endIdx) => {
        // weighted median calculation
        int nRows = endIdx - startIdx + 1; // startIdx and endIdx are inclusive
        if (nRows == 1) return targetArr[idx[startIdx]] - predArr[idx[startIdx]];
        else if (nRows == 2) {
          var w0 = weightArr[idx[startIdx]] * Math.Abs(1.0 / targetArr[idx[startIdx]]);
          var w1 = weightArr[idx[endIdx]] * Math.Abs(1.0 / targetArr[idx[endIdx]]);
          if (w0 > w1) {
            return targetArr[idx[startIdx]] - predArr[idx[startIdx]];
          } else if (w0 < w1) {
            return targetArr[idx[endIdx]] - predArr[idx[endIdx]];
          } else {
            // same weight
            return ((targetArr[idx[startIdx]] - predArr[idx[startIdx]]) + (targetArr[idx[endIdx]] - predArr[idx[endIdx]])) / 2;
          }
        } else {
          var ts = from offset in Enumerable.Range(0, nRows)
                   let i = startIdx + offset
                   select new { res = targetArr[idx[i]] - predArr[idx[i]], weight = weightArr[idx[i]] * Math.Abs(1.0 / targetArr[idx[i]]) };
          ts = ts.OrderBy(t => t.res);
          var totalWeight = ts.Sum(t => t.weight);
          var tsEnumerator = ts.GetEnumerator();
          tsEnumerator.MoveNext();

          double aggWeight = tsEnumerator.Current.weight; // weight of first

          while (aggWeight < totalWeight / 2) {
            tsEnumerator.MoveNext();
            aggWeight += tsEnumerator.Current.weight;
          }
          return tsEnumerator.Current.res;
        }
      };
      return lineSearch;
    }

    public override string ToString() {
      return "Relative error loss";
    }
  }
}
